@@ -15,6 +15,7 @@
 import logging
 import logging.config
 import logging.handlers
+import subprocess
 from copy import deepcopy
 from os import execv, listdir, path, pathsep, stat, system
 from queue import SimpleQueue
@@ -24,11 +25,14 @@ from sys import argv, executable, getsizeof
 from threading import RLock, Thread
 from time import sleep, time
 
+import simplejson
 from simplejson import JSONDecodeError, dumps, load, loads
 from yaml import safe_load
 
 from thingsboard_gateway.gateway.constant_enums import DeviceActions, Status
-from thingsboard_gateway.gateway.constants import CONNECTED_DEVICES_FILENAME, CONNECTOR_PARAMETER, PERSISTENT_GRPC_CONNECTORS_KEY_FILENAME
+from thingsboard_gateway.gateway.constants import CONNECTED_DEVICES_FILENAME, CONNECTOR_PARAMETER, \
+    PERSISTENT_GRPC_CONNECTORS_KEY_FILENAME
+from thingsboard_gateway.gateway.statistics_service import StatisticsService
 from thingsboard_gateway.gateway.tb_client import TBClient
 from thingsboard_gateway.storage.file.file_event_storage import FileEventStorage
 from thingsboard_gateway.storage.memory.memory_event_storage import MemoryEventStorage
@@ -44,6 +48,7 @@ GRPC_LOADED = False
 try:
     from thingsboard_gateway.gateway.grpc_service.grpc_connector import GrpcConnector
     from thingsboard_gateway.gateway.grpc_service.tb_grpc_manager import TBGRPCServerManager
+
     GRPC_LOADED = True
 except ImportError:
     print("Cannot load GRPC connector!")
@@ -55,6 +60,7 @@ DEFAULT_CONNECTORS = {
     "mqtt": "MqttConnector",
     "modbus": "ModbusConnector",
     "opcua": "OpcUaConnector",
+    "opcua_asyncio": "OpcUaConnectorAsyncIO",
     "ble": "BLEConnector",
     "request": "RequestConnector",
     "can": "CanConnector",
@@ -63,8 +69,14 @@ DEFAULT_CONNECTORS = {
     "rest": "RESTConnector",
     "snmp": "SNMPConnector",
     "ftp": "FTPConnector",
-    "socket": "SocketConnector"
-    }
+    "socket": "SocketConnector",
+    "xmpp": "XMPPConnector",
+}
+
+DEFAULT_STATISTIC = {
+    'enable': True,
+    'statsSendPeriodInSeconds': 3600
+}
 
 
 def load_file(path_to_file):
@@ -81,9 +93,10 @@ class TBGatewayService:
         self.async_device_actions = {
             DeviceActions.CONNECT: self.add_device,
             DeviceActions.DISCONNECT: self.del_device
-            }
+        }
         self.__async_device_actions_queue = SimpleQueue()
-        self.__process_async_actions_thread = Thread(target=self.__process_async_device_actions, name="Async device actions processing thread", daemon=True)
+        self.__process_async_actions_thread = Thread(target=self.__process_async_device_actions,
+                                                     name="Async device actions processing thread", daemon=True)
         if config_file is None:
             config_file = path.dirname(path.dirname(path.abspath(__file__))) + '/config/tb_gateway.yaml'.replace('/',
                                                                                                                  path.sep)
@@ -132,14 +145,15 @@ class TBGatewayService:
         self.main_handler.setTarget(self.remote_handler)
         self._default_connectors = DEFAULT_CONNECTORS
         self.__converted_data_queue = SimpleQueue()
-        self.__save_converted_data_thread = Thread(name="Save converted data", daemon=True, target=self.__send_to_storage)
+        self.__save_converted_data_thread = Thread(name="Save converted data", daemon=True,
+                                                   target=self.__send_to_storage)
         self.__save_converted_data_thread.start()
         self._implemented_connectors = {}
         self._event_storage_types = {
             "memory": MemoryEventStorage,
             "file": FileEventStorage,
             "sqlite": SQLiteEventStorage,
-            }
+        }
         self.__gateway_rpc_methods = {
             "ping": self.__rpc_ping,
             "stats": self.__form_statistics,
@@ -148,7 +162,7 @@ class TBGatewayService:
             "version": self.__rpc_version,
             "device_renamed": self.__process_renamed_gateway_devices,
             "device_deleted": self.__process_deleted_gateway_devices,
-            }
+        }
         self.__remote_shell = None
         if self.__config["thingsboard"].get("remoteShell"):
             log.warning("Remote shell is enabled. Please be carefully with this feature.")
@@ -160,8 +174,9 @@ class TBGatewayService:
         self.__rpc_scheduled_methods_functions = {
             "restart": {"function": execv, "arguments": (executable, [executable.split(pathsep)[-1]] + argv)},
             "reboot": {"function": system, "arguments": ("reboot 0",)},
-            }
-        self.__rpc_processing_thread = Thread(target=self.__send_rpc_reply_processing, daemon=True, name="RPC processing thread")
+        }
+        self.__rpc_processing_thread = Thread(target=self.__send_rpc_reply_processing, daemon=True,
+                                              name="RPC processing thread")
         self.__rpc_processing_thread.start()
         self._event_storage = self._event_storage_types[self.__config["storage"]["type"]](self.__config["storage"])
         self.connectors_configs = {}
@@ -186,6 +201,13 @@ class TBGatewayService:
             thread.start()
             log.info('Start checking devices idle time')
 
+        self.__statistics = self.__config['thingsboard'].get('statistics', DEFAULT_STATISTIC)
+        self.__statistics_service = None
+        if self.__statistics['enable'] and self.__statistics.get('configuration'):
+            statistics_config_path = self._config_dir + self.__statistics['configuration']
+            self.__statistics_service = StatisticsService(statistics_config_path,
+                                                          self.__statistics['statsSendPeriodInSeconds'], self, log)
+
         self._published_events = SimpleQueue()
         self._send_thread = Thread(target=self.__read_data_from_storage, daemon=True,
                                    name="Send data to Thingsboard Thread")
@@ -218,7 +240,8 @@ class TBGatewayService:
                                 log.exception(e)
                             if result == 256:
                                 log.warning("Error on RPC command: 256. Permission denied.")
-                if (self.__rpc_requests_in_progress or not self.__rpc_register_queue.empty()) and self.tb_client.is_connected():
+                if (
+                        self.__rpc_requests_in_progress or not self.__rpc_register_queue.empty()) and self.tb_client.is_connected():
                     new_rpc_request_in_progress = {}
                     if self.__rpc_requests_in_progress:
                         for rpc_in_progress, data in self.__rpc_requests_in_progress.items():
@@ -244,8 +267,8 @@ class TBGatewayService:
                     self.__request_config_after_connect = True
                     self.__check_shared_attributes()
 
-                if cur_time - gateway_statistic_send > self.__config["thingsboard"].get("statsSendPeriodInSeconds",
-                                                                                        3600) * 1000 and self.tb_client.is_connected():
+                if cur_time - gateway_statistic_send > self.__statistics[
+                        'statsSendPeriodInSeconds'] * 1000 and self.tb_client.is_connected():
                     summary_messages = self.__form_statistics()
                     # with self.__lock:
                     self.tb_client.client.send_telemetry(summary_messages)
@@ -282,6 +305,10 @@ class TBGatewayService:
         self.stopped = True
         self.__updater.stop()
         log.info("Stopping...")
+
+        if self.__statistics_service:
+            self.__statistics_service.stop()
+
         if self.__grpc_manager is not None:
             self.__grpc_manager.stop()
         self.__close_connectors()
@@ -387,11 +414,17 @@ class TBGatewayService:
         self.tb_client.client.subscribe_to_all_attributes(self._attribute_update_callback)
         self.tb_client.client.gw_subscribe_to_all_attributes(self._attribute_update_callback)
 
+    def request_device_attributes(self, device_name, shared_keys, client_keys, callback):
+        if client_keys is not None:
+            self.tb_client.client.gw_request_client_attributes(device_name, client_keys, callback)
+        if shared_keys is not None:
+            self.tb_client.client.gw_request_shared_attributes(device_name, shared_keys, callback)
+
     def __check_shared_attributes(self):
         self.tb_client.client.request_attributes(callback=self._attributes_parse)
-
     def __register_connector(self, session_id, connector_key):
-        if self.__grpc_connectors.get(connector_key) is not None and self.__grpc_connectors[connector_key]['name'] not in self.available_connectors:
+        if self.__grpc_connectors.get(connector_key) is not None and self.__grpc_connectors[connector_key][
+                'name'] not in self.available_connectors:
             target_connector = self.__grpc_connectors.get(connector_key)
             connector = GrpcConnector(self, target_connector['config'], self.__grpc_manager, session_id)
             connector.setName(target_connector['name'])
@@ -406,17 +439,30 @@ class TBGatewayService:
             log.error("GRPC configuration for connector with key: %s - not found", connector_key)
 
     def __unregister_connector(self, session_id, connector_key):
-        if self.__grpc_connectors.get(connector_key) is not None and self.__grpc_connectors[connector_key]['name'] in self.available_connectors:
+        if self.__grpc_connectors.get(connector_key) is not None and self.__grpc_connectors[connector_key][
+                'name'] in self.available_connectors:
             connector_name = self.__grpc_connectors[connector_key]['name']
             target_connector: GrpcConnector = self.available_connectors.pop(connector_name)
             self.__grpc_manager.unregister(Status.SUCCESS, session_id, target_connector)
-            log.info("GRPC connector with key %s and name %s - unregistered", connector_key, target_connector.get_name())
+            log.info("GRPC connector with key %s and name %s - unregistered", connector_key,
+                     target_connector.get_name())
         elif self.__grpc_connectors.get(connector_key) is not None:
             self.__grpc_manager.unregister(Status.NOT_FOUND, session_id, None)
             log.error("GRPC connector with key: %s - is not registered!", connector_key)
         else:
             self.__grpc_manager.unregister(Status.FAILURE, session_id, None)
-            log.error("GRPC configuration for connector with key: %s - not found in configuration and not registered", connector_key)
+            log.error("GRPC configuration for connector with key: %s - not found in configuration and not registered",
+                      connector_key)
+
+    @staticmethod
+    def _generate_persistent_key(connector, connectors_persistent_keys):
+        if connectors_persistent_keys and connectors_persistent_keys.get(connector['name']) is not None:
+            connector_persistent_key = connectors_persistent_keys[connector['name']]
+        else:
+            connector_persistent_key = "".join(choice(hexdigits) for _ in range(10))
+            connectors_persistent_keys[connector['name']] = connector_persistent_key
+
+        return connector_persistent_key
 
     def _load_connectors(self):
         self.connectors_configs = {}
@@ -432,21 +478,33 @@ class TBGatewayService:
                     connector_pass = connector.get("pass", None)
 
                     if connector['type'] == "grpc" and self.__grpc_manager is None:
-                        log.error("Cannot load connector with name: %s and type grpc. GRPC server is disabled!", connector['name'])
+                        log.error("Cannot load connector with name: %s and type grpc. GRPC server is disabled!",
+                                  connector['name'])
                         continue
+
                     if connector['type'] != "grpc":
-                        connector_class = TBModuleLoader.import_module(connector['type'], self._default_connectors.get(connector['type'], connector.get('class')))
+                        connector_class = None
+                        if connector.get('useGRPC', True):
+                            module_name = f'Grpc{self._default_connectors.get(connector["type"], connector.get("class"))}'
+                            connector_class = TBModuleLoader.import_module(connector['type'], module_name)
+
+                        if self.__grpc_manager and self.__grpc_manager.is_alive() and connector_class:
+                            connector_persistent_key = self._generate_persistent_key(connector,
+                                                                                     connectors_persistent_keys)
+                        else:
+                            connector_class = TBModuleLoader.import_module(connector['type'],
+                                                                           self._default_connectors.get(
+                                                                               connector['type'],
+                                                                               connector.get('class')))
+
                         self._implemented_connectors[connector['type']] = connector_class
                     elif connector['type'] == "grpc":
                         if connector.get('key') == "auto":
-                            if connectors_persistent_keys and connectors_persistent_keys.get(connector['name']) is not None:
-                                connector_persistent_key = connectors_persistent_keys[connector['name']]
-                            else:
-                                connector_persistent_key = "".join(choice(hexdigits) for _ in range(10))
-                                connectors_persistent_keys[connector['name']] = connector_persistent_key
+                            self._generate_persistent_key(connector, connectors_persistent_keys)
                         else:
                             connector_persistent_key = connector['key']
-                        log.info("Connector key for GRPC connector with name [%s] is: [%s]", connector['name'], connector_persistent_key)
+                        log.info("Connector key for GRPC connector with name [%s] is: [%s]", connector['name'],
+                                 connector_persistent_key)
                     config_file_path = self._config_dir + connector['configuration']
                     connector_conf_file_data = ''
                     with open(config_file_path, 'r', encoding="UTF-8") as conf_file:
@@ -465,7 +523,10 @@ class TBGatewayService:
                         connector_conf["name"] = connector['name']
                     #TO DO
                     self.connectors_configs[connector['type']].append({"name": connector['name'],
-                                                                       "config": {connector['configuration']: connector_conf} if connector['type'] != 'grpc' else connector_conf,
+                                                                       "config": {connector[
+                                                                                      'configuration']: connector_conf} if
+                                                                       connector[
+                                                                           'type'] != 'grpc' else connector_conf,
                                                                        "config_updated": stat(config_file_path),
                                                                        "config_file_path": config_file_path,
                                                                        "grpc_key": connector_persistent_key,
@@ -485,7 +546,8 @@ class TBGatewayService:
     def _connect_with_connectors(self):
         for connector_type in self.connectors_configs:
             for connector_config in self.connectors_configs[connector_type]:
-                if connector_type.lower() != 'grpc':
+                if connector_type.lower() != 'grpc' and (self._implemented_connectors.get(connector_type.lower()) is not None and
+                        'Grpc' not in self._implemented_connectors[connector_type.lower()].__name__):
                     hostIP = connector_config.get("hostIP", None)
                     clientID = connector_config.get("clientID", None)
                     secUser = connector_config.get("user", None)
@@ -518,6 +580,26 @@ class TBGatewayService:
                                 connector.close()
                 else:
                     self.__grpc_connectors.update({connector_config['grpc_key']: connector_config})
+                    if connector_type.lower() != 'grpc':
+                        connector_dir_abs = "/".join(self._config_dir.split("/")[:-2])
+                        connector_file_name = f'{connector_type}_connector.py'
+                        connector_abs_path = f'{connector_dir_abs}/grpc_connectors/{connector_type}/{connector_file_name}'
+                        connector_config_json = simplejson.dumps({
+                            **connector_config,
+                            'gateway': {
+                                'host': 'localhost',
+                                'port': self.__config['grpc']['serverPort']
+                            }
+                        })
+
+                        thread = Thread(target=self._run_connector, args=(connector_abs_path, connector_config_json,),
+                                        daemon=True, name='Separate DRPC Connector')
+                        thread.start()
+
+    def _run_connector(self, connector_abs_path, connector_config_json):
+        subprocess.run(['python3', connector_abs_path, connector_config_json, self._config_dir],
+                       check=True,
+                       universal_newlines=True)
 
     def check_connector_configuration_updates(self):
         configuration_changed = False
@@ -583,30 +665,49 @@ class TBGatewayService:
 
                         max_data_size = self.__config["thingsboard"].get("maxPayloadSizeBytes", 1024)
                         if self.__get_data_size(data) >= max_data_size:
+                            # Data is too large, so we will attempt to send in pieces
                             adopted_data = {"deviceName": data['deviceName'],
                                             "deviceType": data['deviceType'],
                                             "attributes": {},
                                             "telemetry": []}
+
+                            # First, loop through the attributes
                             for attribute in data['attributes']:
+                                adopted_data['attributes'].update(attribute)
                                 adopted_data_size = self.__get_data_size(adopted_data)
                                 if adopted_data_size >= max_data_size:
+                                    # We have surpassed the max_data_size, so send what we have and clear attributes
                                     self.__send_data_pack_to_storage(adopted_data, connector_name)
                                     adopted_data['attributes'] = {}
-                                adopted_data['attributes'].update({attribute: data['attributes'][attribute]})
-                            for ts_kv_list in data['telemetry']:
+                            
+                            # Now, loop through telemetry. Possibly have some unsent attributes that have been adopted.
+                            telemetry = data['telemetry'] if isinstance(data['telemetry'], list) else [data['telemetry']]
+                            for ts_kv_list in telemetry:
                                 ts = ts_kv_list['ts']
                                 for kv in ts_kv_list['values']:
-                                    adopted_data_size = self.__get_data_size(adopted_data)
-                                    if adopted_data_size >= max_data_size:
-                                        self.__send_data_pack_to_storage(adopted_data, connector_name)
-                                        adopted_data['telemetry'] = []
                                     if len(adopted_data['telemetry']) == 0:
-                                        adopted_data['telemetry'] = [{'ts': ts, 'values': {kv: ts_kv_list['values'][kv]}}]
+                                        adopted_data['telemetry'] = [
+                                            {'ts': ts, 'values': {kv: ts_kv_list['values'][kv]}}]
                                     else:
                                         for adopted_kv in adopted_data['telemetry']:
                                             if adopted_kv['ts'] == ts:
                                                 adopted_kv['values'].update({kv: ts_kv_list['values'][kv]})
 
+                                    adopted_data_size = self.__get_data_size(adopted_data)
+                                    if adopted_data_size >= max_data_size:
+                                        # we have surpassed the max_data_size, so send what we have and clear attributes and telemetry
+                                        self.__send_data_pack_to_storage(adopted_data, connector_name)
+                                        adopted_data['telemetry'] = []
+                                        adopted_data['attributes'] = {}
+
+                            # It is possible that we get here and have some telemetry or attributes not yet sent, so check for that.
+                            if len(adopted_data['telemetry']) > 0 or len(adopted_data['attributes']) > 0:
+                                self.__send_data_pack_to_storage(adopted_data, connector_name)
+                                
+                                # technically unnecessary to clear here, but leaving for consistency.
+                                adopted_data['telemetry'] = []
+                                adopted_data['attributes'] = {}
+
                         else:
                             self.__send_data_pack_to_storage(data, connector_name)
 
@@ -645,7 +746,8 @@ class TBGatewayService:
                       connector_name)
 
     def check_size(self, devices_data_in_event_pack):
-        if self.__get_data_size(devices_data_in_event_pack) >= self.__config["thingsboard"].get("maxPayloadSizeBytes", 1024):
+        if self.__get_data_size(devices_data_in_event_pack) >= self.__config["thingsboard"].get("maxPayloadSizeBytes",
+                                                                                                1024):
             self.__send_data(devices_data_in_event_pack)
             for device in devices_data_in_event_pack:
                 devices_data_in_event_pack[device]["telemetry"] = []
@@ -896,7 +998,8 @@ class TBGatewayService:
             else:
                 sleep(.1)
 
-    def __send_rpc_reply(self, device=None, req_id=None, content=None, success_sent=None, wait_for_publish=None, quality_of_service=0):
+    def __send_rpc_reply(self, device=None, req_id=None, content=None, success_sent=None, wait_for_publish=None,
+                         quality_of_service=0):
         try:
             self.__rpc_reply_sent = True
             rpc_response = {"success": False}
@@ -988,15 +1091,18 @@ class TBGatewayService:
         self.__saved_devices.pop(device_name)
         self.__save_persistent_devices()
 
-    def get_devices(self):
-        return self.__connected_devices
+    def get_devices(self, connector_name: str = None):
+        return self.__connected_devices if connector_name is None else {device_name: self.__connected_devices[device_name]["device_type"] for device_name in self.__connected_devices.keys() if self.__connected_devices[device_name].get("connector") is not None and
+                                                                           self.__connected_devices[device_name]["connector"].get_name() == connector_name}
+
 
     def __process_async_device_actions(self):
         while not self.stopped:
             if not self.__async_device_actions_queue.empty():
                 action, data = self.__async_device_actions_queue.get()
                 if action == DeviceActions.CONNECT:
-                    self.add_device(data['deviceName'], {CONNECTOR_PARAMETER: self.available_connectors[data['name']]}, data.get('deviceType'))
+                    self.add_device(data['deviceName'], {CONNECTOR_PARAMETER: self.available_connectors[data['name']]},
+                                    data.get('deviceType'))
                 elif action == DeviceActions.DISCONNECT:
                     self.del_device(data['deviceName'])
             else:
